@@ -6,6 +6,9 @@ import time
 
 
 class Effect(metaclass=abc.ABCMeta):
+    """
+    Base class for effects which can be yielded to the tapystry event loop.
+    """
     def __init__(self, type, oncancel=(lambda: None), name=None):
         self.type = type
         self.cancel = oncancel
@@ -18,6 +21,9 @@ class Effect(metaclass=abc.ABCMeta):
 
 
 class Broadcast(Effect):
+    """
+    Effect which broadcasts a message for all strands to hear
+    """
     def __init__(self, key, value=None, name=None, immediate=False, **effect_kwargs):
         self.key = key
         self.value = value
@@ -28,6 +34,10 @@ class Broadcast(Effect):
 
 
 class Receive(Effect):
+    """
+    Effect which waits until it hears a broadcast at the specified key, with value satisfying the specified predicate.
+    The tapystry engine returns the matched message's value
+    """
     def __init__(self, key, predicate=None, name=None, **effect_kwargs):
         self.key = key
         self.predicate = predicate
@@ -37,6 +47,10 @@ class Receive(Effect):
 
 
 class Call(Effect):
+    """
+    Effect which spins up a new strand by calling generator on the specified arguments,
+    The tapystry engine returns the generator's return value
+    """
     def __init__(self, gen, args=(), kwargs=None, name=None, **effect_kwargs):
         self.gen = gen
         self.args = args
@@ -47,6 +61,10 @@ class Call(Effect):
 
 
 class CallFork(Effect):
+    """
+    Effect which spins up a new strand by calling generator on the specified arguments
+    The tapystry engine immediately returns a Strand object.
+    """
     def __init__(self, gen, args=(), kwargs=None, name=None, **effect_kwargs):
         self.gen = gen
         self.args = args
@@ -57,7 +75,12 @@ class CallFork(Effect):
 
 
 class First(Effect):
-    """NOTE: use of this can be dangerous, as it cancels losers"""
+    """
+    Effect which returns when one of the strands is done.
+    The tapystry engine returns the index of the winning strand, and its value.
+    NOTE: Use of this can be dangerous and can lead to deadlocks, as it cancels losers.
+          It is safer to us higher-level APIs such as Race and Join
+    """
     def __init__(self, strands, name=None, cancel_losers=True, **effect_kwargs):
         self.strands = strands
         self.cancel_losers = cancel_losers
@@ -69,6 +92,9 @@ class First(Effect):
 
 # TODO: does this really need to be an effect?  what's wrong with just exposing _canceled on Strand?
 class Cancel(Effect):
+    """
+    Effect which cancels the strand specified
+    """
     def __init__(self, strand, name=None, **effect_kwargs):
         self.strand = strand
         if name is None:
@@ -77,6 +103,9 @@ class Cancel(Effect):
 
 
 class Sleep(Effect):
+    """
+    Effect which sleeps the calling strand for the specified period of time
+    """
     def __init__(self, t, name=None, **effect_kwargs):
         self.t = t
         if name is None:
@@ -84,6 +113,20 @@ class Sleep(Effect):
         super().__init__(type="Sleep", name=name, **effect_kwargs)
 
 
+class Intercept(Effect):
+    """
+    Effect which waits until the engine finds an effect matching the given predicate, and modifies the yielded value of that effect.
+    This is intended for testing only, and can only be used in test_mode.
+    The tapystry engine returns the effect found.
+    """
+    def __init__(self, predicate=None, value=None, name=None, **effect_kwargs):
+        self.predicate = predicate
+        self.value = value
+        if name is None:
+            name = ""
+        super().__init__(type="Intercept", name=name, **effect_kwargs)
+
+
 class TapystryError(Exception):
     pass
 
@@ -171,13 +214,17 @@ class _QueueItem():
         self.wake_time = wake_time
 
 
-def run(gen, args=(), kwargs=None, debug=False):
+def run(gen, args=(), kwargs=None, debug=False, test_mode=False):
     # dict from string to waiting functions
     waiting = defaultdict(list)
     # dict from strand to waiting key
     # TODO: gc hanging strands
     hanging_strands = set()
     q = deque()
+
+    # list of intercept items
+    intercepts = []
+
     initial_strand = Strand(gen, args, kwargs)
     if initial_strand.is_done():
         # wasn't even a generator
@@ -263,6 +310,28 @@ def run(gen, args=(), kwargs=None, debug=False):
     while len(q):
         item = q.pop()
 
+        effect = item.effect
+        if isinstance(effect, Intercept):
+            if not test_mode:
+                raise TapystryError(f"Cannot intercept outside of test mode!")
+            intercepts.append(item)
+            hanging_strands.add(item.strand)
+            continue
+
+        if test_mode:
+            intercepted = False
+            for intercept_item in intercepts:
+                intercept_effect = intercept_item.effect
+                if intercept_effect.predicate is None or intercept_effect.predicate(effect):
+                    intercepted = True
+                    break
+            if intercepted:
+                hanging_strands.remove(intercept_item.strand)
+                advance_strand(intercept_item.strand, effect)
+                advance_strand(item.strand, intercept_effect.value)
+                intercepts.remove(intercept_item)
+                continue
+
         if item.wake_time is not None and item.wake_time > time.time():
             q.appendleft(item)
 
@@ -276,7 +345,7 @@ def run(gen, args=(), kwargs=None, debug=False):
 
         if item.strand.is_canceled():
             continue
-        effect = item.effect
+
         if debug:
             print(f"Handling {effect} (from {item.strand})")
 
